import unittest

from pyais.ais_types import AISType
from pyais.constants import ManeuverIndicator, NavigationStatus, ShipType, NavAid, EpfdType
from pyais.exceptions import UnknownMessageException
from pyais.messages import AISMessage, NMEAMessage
from pyais.stream import ByteStream


class TestAIS(unittest.TestCase):
    """
    TestCases for AIS message decoding and assembling.

    The Test messages are from multiple sources and are scrambled together.
    Raw messages are decoded by either hand or some online decoder.
    As my main source of AIS messages I used this dumb:
    https://www.aishub.net/ais-dispatcher

    As my main decoder I used this decoder:
    http://ais.tbsalling.dk

    The latter sometimes is a bit weird and therefore I used aislib to verify my results.
    """

    def test_nmea(self):
        """
        Test if ais message still contains the original nmea message
        """
        nmea = NMEAMessage(b"!AIVDM,1,1,,B,91b55wi;hbOS@OdQAC062Ch2089h,0*30")
        assert nmea.decode().nmea == nmea

    def test_to_json(self):
        json_dump = NMEAMessage(b"!AIVDM,1,1,,A,15NPOOPP00o?b=bE`UNv4?w428D;,0*24").decode().to_json()
        text = """{
    "nmea": {
        "ais_id": 1,
        "raw": "!AIVDM,1,1,,A,15NPOOPP00o?b=bE`UNv4?w428D;,0*24",
        "talker": "AI",
        "type": "VDM",
        "message_fragments": 1,
        "fragment_number": 1,
        "message_id": null,
        "channel": "A",
        "payload": "15NPOOPP00o?b=bE`UNv4?w428D;",
        "fill_bits": 0,
        "checksum": 36,
        "bit_array": "000001000101011110100000011111011111100000100000000000000000110111001111101010001101101010010101101000100101011110111110000100001111111111000100000010001000010100001011"
    },
    "decoded": {
        "type": 1,
        "repeat": 0,
        "mmsi": "367533950",
        "status": 0,
        "turn": -128,
        "speed": 0.0,
        "accuracy": 1,
        "lon": -122.40823166666667,
        "lat": 37.808418333333336,
        "course": 360.0,
        "heading": 511,
        "second": 34,
        "maneuver": 0,
        "raim": 1,
        "radio": 34059
    }
}"""
        self.assertEqual(json_dump, text)

    def test_msg_type(self):
        """
        Test if msg type is correct
        """
        nmea = NMEAMessage(b"!AIVDM,1,1,,B,15M67FC000G?ufbE`FepT@3n00Sa,0*5C")
        assert nmea.decode().msg_type == AISType.POS_CLASS_A1

        nmea = NMEAMessage(b"!AIVDM,1,1,,B,15NG6V0P01G?cFhE`R2IU?wn28R>,0*05")
        assert nmea.decode().msg_type == AISType.POS_CLASS_A1

        nmea = NMEAMessage.assemble_from_iterable(messages=[
            NMEAMessage(b"!AIVDM,2,1,4,A,55O0W7`00001L@gCWGA2uItLth@DqtL5@F22220j1h742t0Ht0000000,0*08"),
            NMEAMessage(b"!AIVDM,2,2,4,A,000000000000000,2*20")
        ])
        assert nmea.decode().msg_type == AISType.STATIC_AND_VOYAGE

    def test_msg_getitem(self):
        """
        Test if one can get items
        """
        msg = NMEAMessage(b"!AIVDM,1,1,,B,15NG6V0P01G?cFhE`R2IU?wn28R>,0*05").decode()
        assert msg['repeat'] == 0

    def test_msg_type_1(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,15M67FC000G?ufbE`FepT@3n00Sa,0*5C").decode()
        assert msg.content == {'type': 1, 'repeat': 0, 'mmsi': "366053209",
                               'status': NavigationStatus.RestrictedManoeuverability, 'turn': 0,
                               'speed': 0,
                               'accuracy': 0,
                               'lon': -122.34161833333333, 'lat': 37.80211833333333, 'course': 219.3,
                               'heading': 1,
                               'second': 59, 'maneuver': ManeuverIndicator.NotAvailable, 'raim': False,
                               'radio': 2281}

        msg = NMEAMessage(b"!AIVDM,1,1,,A,15NPOOPP00o?b=bE`UNv4?w428D;,0*24").decode()
        assert msg['type'] == 1
        assert msg['mmsi'] == "367533950"
        assert msg['repeat'] == 0
        assert msg['status'] == NavigationStatus.UnderWayUsingEngine
        assert msg['turn'] == -128
        assert msg['speed'] == 0
        assert msg['accuracy'] == 1
        assert round(msg['lat'], 4) == 37.8084
        assert round(msg['lon'], 4) == -122.4082
        assert msg['course'] == 360
        assert msg['heading'] == 511
        assert msg['second'] == 34
        assert msg['maneuver'] == ManeuverIndicator.NotAvailable
        assert msg['raim']

        msg = NMEAMessage(b"!AIVDM,1,1,,B,181:Kjh01ewHFRPDK1s3IRcn06sd,0*08").decode()
        assert msg['course'] == 87.0
        assert msg['mmsi'] == "538090443"
        assert msg['speed'] == 10.9

    def test_decode_pos_1_2_3(self):
        # weired message of type 0 as part of issue #4
        msg: NMEAMessage = NMEAMessage(b"!AIVDM,1,1,,B,0S9edj0P03PecbBN`ja@0?w42cFC,0*7C")

        assert msg.is_valid
        content: AISMessage = msg.decode(silent=False)
        assert msg

        assert content['repeat'] == 2
        assert content['mmsi'] == "211512520"
        assert content['turn'] == -128
        assert content['speed'] == 0.3
        assert round(content['lat'], 4) == 53.5427
        assert round(content['lon'], 4) == 9.9794
        assert round(content['course'], 1) == 0.0

        msg: NMEAMessage = NMEAMessage(b"!AIVDM,1,1,,B,0S9edj0P03PecbBN`ja@0?w42cFC,0*7C")
        assert msg.decode().to_json()

    def test_msg_type_3(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,35NSH95001G?wopE`beasVk@0E5:,0*6F").decode()
        assert msg['type'] == 3
        assert msg['mmsi'] == "367581220"
        assert msg['repeat'] == 0
        assert msg['status'] == NavigationStatus.Moored
        assert msg['turn'] == 0
        assert msg['speed'] == 0.1
        assert msg['accuracy'] == 0
        assert round(msg['lat'], 4) == 37.8107
        assert round(msg['lon'], 4) == -122.3343
        assert round(msg['course'], 1) == 254.2
        assert msg['heading'] == 217
        assert msg['second'] == 40
        assert msg['maneuver'] == ManeuverIndicator.NotAvailable
        assert not msg['raim']

    def test_msg_type_4(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,403OviQuMGCqWrRO9>E6fE700@GO,0*4D").decode()
        assert round(msg['lon'], 4) == -76.3524
        assert round(msg['lat'], 4) == 36.8838
        assert msg['accuracy'] == 1
        assert msg['year'] == 2007
        assert msg['month'] == 5
        assert msg['day'] == 14
        assert msg['minute'] == 57
        assert msg['second'] == 39

        msg = NMEAMessage(b"!AIVDM,1,1,,B,403OtVAv>lba;o?Ia`E`4G?02H6k,0*44").decode()
        assert round(msg['lon'], 4) == -122.4648
        assert round(msg['lat'], 4) == 37.7943
        assert msg['mmsi'] == "003669145"
        assert msg['accuracy'] == 1
        assert msg['year'] == 2019
        assert msg['month'] == 11
        assert msg['day'] == 9
        assert msg['hour'] == 10
        assert msg['minute'] == 41
        assert msg['second'] == 11

    def test_msg_type_5(self):
        msg = NMEAMessage.assemble_from_iterable(messages=[
            NMEAMessage(b"!AIVDM,2,1,1,A,55?MbV02;H;s<HtKR20EHE:0@T4@Dn2222222216L961O5Gf0NSQEp6ClRp8,0*1C"),
            NMEAMessage(b"!AIVDM,2,2,1,A,88888888880,2*25")
        ]).decode()
        assert msg['callsign'] == "3FOF8"
        assert msg['shipname'] == "EVER DIADEM"
        assert msg['shiptype'] == ShipType.Cargo
        assert msg['to_bow'] == 225
        assert msg['to_stern'] == 70
        assert msg['to_port'] == 1
        assert msg['to_starboard'] == 31
        assert msg['draught'] == 12.2
        assert msg['destination'] == "NEW YORK"
        assert msg['dte'] == 0

    def test_msg_type_6(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,6B?n;be:cbapalgc;i6?Ow4,2*4A").decode()
        assert msg['seqno'] == 3
        assert msg['dest_mmsi'] == "313240222"
        assert msg['dac'] == 669
        assert msg['fid'] == 11

    def test_msg_type_7(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,702R5`hwCjq8,0*6B").decode()
        assert msg['mmsi1'] == "265538450"

    def test_msg_type_8(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,85Mwp`1Kf3aCnsNvBWLi=wQuNhA5t43N`5nCuI=p<IBfVqnMgPGs,0*47").decode()

        assert msg.nmea.is_valid
        assert msg['repeat'] == 0
        assert msg.msg_type == AISType.BINARY_BROADCAST
        assert msg['mmsi'] == "366999712"
        assert msg['dac'] == 366
        assert msg['fid'] == 56
        assert msg['data'] == "00111010010100111101101110110111101111100100101001110111001100010011011111111" \
                              "00001111101011110110000010001000101111100000100000011011110101000000101110110" \
                              "01001111110101100100110111100000110001100101001010111010011011100111011001110" \
                              "1101111100000010111111011"

    def test_msg_type_9(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,91b55wi;hbOS@OdQAC062Ch2089h,0*30").decode()
        assert msg.msg_type == AISType.SAR_AIRCRAFT_POS
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "111232511"
        assert msg['alt'] == 303
        assert msg['speed'] == 42
        assert msg['accuracy'] == 0
        assert round(msg['lon'], 5) == -6.27884
        assert round(msg['lat'], 5) == 58.144
        assert msg['course'] == 154.5
        assert msg['second'] == 15
        assert msg['dte'] == 1
        assert msg['radio'] == 33392

    def test_msg_type_10(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,:5MlU41GMK6@,0*6C").decode()
        assert msg['dest_mmsi'] == "366832740"

        msg = NMEAMessage(b"!AIVDM,1,1,,B,:6TMCD1GOS60,0*5B").decode()
        assert msg['dest_mmsi'] == "366972000"

    def test_msg_type_11(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,;4R33:1uUK2F`q?mOt@@GoQ00000,0*5D").decode()
        assert round(msg['lon'], 4) == -94.4077
        assert round(msg['lat'], 4) == 28.4091
        assert msg['accuracy'] == 1
        assert msg['type'] == 11
        assert msg['year'] == 2009
        assert msg['month'] == 5
        assert msg['day'] == 22
        assert msg['hour'] == 2
        assert msg['minute'] == 22
        assert msg['second'] == 40

    def test_msg_type_12(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,<5?SIj1;GbD07??4,0*38").decode()
        assert msg['type'] == 12
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "351853000"
        assert msg['seqno'] == 0
        assert msg['dest_mmsi'] == "316123456"
        assert msg['retransmit'] == 0
        assert msg['text'] == "GOOD"

        msg = NMEAMessage(b"!AIVDM,1,1,,A,<42Lati0W:Ov=C7P6B?=Pjoihhjhqq0,2*2B").decode()
        assert msg['type'] == 12
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "271002099"
        assert msg['seqno'] == 0
        assert msg['dest_mmsi'] == "271002111"
        assert msg['retransmit'] == 1
        assert msg['text'] == "MSG FROM 271002099"

    def test_msg_type_13(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,=39UOj0jFs9R,0*65").decode()
        assert msg['type'] == 13
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "211378120"
        assert msg['mmsi1'] == "211217560"

    def test_msg_type_14(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,>5?Per18=HB1U:1@E=B0m<L,2*51").decode()
        assert msg['type'] == 14
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "351809000"
        assert msg['text'] == "RCVD YR TEST MSG"

    def test_msg_type_15(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,?5OP=l00052HD00,2*5B").decode()
        assert msg['type'] == 15
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "368578000"
        assert msg['offset1_1'] == 0

        msg = NMEAMessage(b"!AIVDM,1,1,,B,?h3Ovn1GP<K0<P@59a0,2*04").decode()
        assert msg['type'] == 15
        assert msg['repeat'] == 3
        assert msg['mmsi'] == "003669720"
        assert msg['mmsi1'] == "367014320"
        assert msg['type1_1'] == 3
        assert msg['type1_2'] == 5
        assert msg['offset1_2'] == 617
        assert msg['offset1_1'] == 516

    def test_msg_type_16(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,@01uEO@mMk7P<P00,0*18").decode()
        assert msg['type'] == 16
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "002053501"
        assert msg['mmsi1'] == "224251000"
        assert msg['offset1'] == 200
        assert msg['increment1'] == 0

        assert msg['offset2'] == 0
        assert msg['increment1'] == 0

    def test_msg_type_17(self):
        msg = NMEAMessage.assemble_from_iterable(messages=[
            NMEAMessage(b"!AIVDM,2,1,5,A,A02VqLPA4I6C07h5Ed1h<OrsuBTTwS?r:C?w`?la<gno1RTRwSP9:BcurA8a,0*3A"),
            NMEAMessage(b"!AIVDM,2,2,5,A,:Oko02TSwu8<:Jbb,0*11")
        ]).decode()
        n = 0x7c0556c07031febbf52924fe33fa2933ffa0fd2932fdb7062922fe3809292afde9122929fcf7002923ffd20c29aaaa
        assert msg['type'] == 17
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "002734450"
        assert msg['lon'] == 17478
        assert msg['lat'] == 35992
        assert msg['data'] == n

        msg = NMEAMessage(b"!AIVDM,1,1,,A,A0476BQ>J8`<h2JpH:4P0?j@2mTEw8`=DP1DEnqvj0,0*79").decode()
        assert msg['type'] == 17
        assert msg['repeat'] == 0
        assert msg['mmsi'] == "004310602"
        assert msg['lat'] == 20582
        assert msg['lon'] == 80290
        assert msg['data'] == 14486955885545814640451754168044205828166539334830080

    def test_msg_type_18(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,B5NJ;PP005l4ot5Isbl03wsUkP06,0*76").decode()
        assert msg['type'] == 18
        assert msg['mmsi'] == "367430530"
        assert msg['speed'] == 0
        assert msg['accuracy'] == 0
        assert round(msg['lat'], 2) == 37.79
        assert round(msg['lon'], 2) == -122.27
        assert msg['course'] == 0
        assert msg['heading'] == 511
        assert msg['second'] == 55
        assert msg['regional'] == 0
        assert msg['cs'] == 1
        assert msg['display'] == 0
        assert msg['dsc'] == 1
        assert msg['msg22'] == 1
        assert msg['raim'] == 0

    def test_msg_type_19(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,C5N3SRgPEnJGEBT>NhWAwwo862PaLELTBJ:V00000000S0D:R220,0*0B").decode()
        assert msg['type'] == 19
        assert msg['mmsi'] == "367059850"
        assert round(msg['speed'], 1) == 8.7
        assert msg['accuracy'] == 0
        assert round(msg['lat'], 2) == 29.54
        assert round(msg['lon'], 2) == -88.81
        assert round(msg['course'], 2) == 335.9
        assert msg['heading'] == 511
        assert msg['second'] == 46
        assert msg['shipname'] == "CAPT.J.RIMES"
        assert msg['shiptype'] == ShipType(70)
        assert msg['to_bow'] == 5
        assert msg['to_stern'] == 21
        assert msg['to_port'] == 4
        assert msg['to_starboard'] == 4

    def test_msg_type_20(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,D028rqP<QNfp000000000000000,2*0C").decode()
        assert msg['type'] == 20
        assert msg['mmsi'] == "002243302"
        assert msg['offset1'] == 200
        assert msg['number1'] == 5
        assert msg['timeout1'] == 7
        assert msg['increment1'] == 750

        # All other values are zero
        for k, v in msg.content.items():
            if k not in ('type', 'mmsi', 'offset1', 'number1', 'timeout1', 'increment1'):
                assert not v

    def test_msg_type_21(self):
        msg = NMEAMessage.assemble_from_iterable(messages=[
            NMEAMessage(b"!AIVDM,2,1,7,B,E4eHJhPR37q0000000000000000KUOSc=rq4h00000a,0*4A"),
            NMEAMessage(b"!AIVDM,2,2,7,B,@20,4*54")
        ]).decode()
        assert msg['type'] == 21
        assert msg['mmsi'] == "316021442"
        assert msg['aid_type'] == NavAid.REFERENCE_POINT
        assert msg['name'] == "DFO2"
        assert msg['accuracy'] == 1
        assert round(msg['lat'], 2) == 48.65
        assert round(msg['lon'], 2) == -123.43
        assert not msg['to_bow']
        assert not msg['to_stern']
        assert not msg['to_port']
        assert not msg['to_starboard']

        assert msg['off_position']
        assert msg['regional'] == 0
        assert msg['raim']
        assert msg['virtual_aid'] == 0
        assert msg['assigned'] == 0
        assert msg['name_extension'] == ""

    def test_msg_type_22(self):
        # Broadcast
        msg = NMEAMessage(b"!AIVDM,1,1,,B,F030p:j2N2P5aJR0r;6f3rj10000,0*11").decode()
        assert msg['type'] == 22
        assert msg['mmsi'] == "003160107"
        assert msg['channel_a'] == 2087
        assert msg['channel_b'] == 2088
        assert msg['power'] == 0

        assert msg['ne_lon'] == -7710.0
        assert msg['ne_lat'] == 3300.0
        assert msg['sw_lon'] == -8020.0
        assert msg['sw_lat'] == 3210

        assert msg['band_a'] == 0
        assert msg['band_b'] == 0
        assert msg['zonesize'] == 2

        assert 'dest1' not in msg.content.keys()
        assert 'dest2' not in msg.content.keys()

        # Addressed
        msg = NMEAMessage(b"!AIVDM,1,1,,A,F@@W>gOP00PH=JrN9l000?wB2HH;,0*44").decode()
        assert msg['type'] == 22
        assert msg['mmsi'] == "017419965"
        assert msg['channel_a'] == 3584
        assert msg['channel_b'] == 8
        assert msg['power'] == 1
        assert msg['addressed'] == 1

        assert msg['dest1'] == "028144881"
        assert msg['dest2'] == "268435519"

        assert msg['band_a'] == 0
        assert msg['band_b'] == 0
        assert msg['zonesize'] == 4

        assert 'ne_lon' not in msg.content.keys()
        assert 'ne_lat' not in msg.content.keys()
        assert 'sw_lon' not in msg.content.keys()
        assert 'sw_lat' not in msg.content.keys()

    def test_msg_type_23(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,G02:Kn01R`sn@291nj600000900,2*12").decode()
        assert msg['type'] == 23
        assert msg['mmsi'] == "002268120"
        assert msg['ne_lon'] == 157.8
        assert round(msg['ne_lat'], 1) == 3064.2
        assert round(msg['sw_lon'], 1) == 109.6
        assert round(msg['sw_lat'], 1) == 3040.8

    def test_msg_type_24(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,H52KMeDU653hhhi0000000000000,0*1A").decode()
        assert msg['type'] == 24
        assert msg['mmsi'] == "338091445"
        assert msg['partno'] == 1
        assert msg['shiptype'] == ShipType.PleasureCraft
        assert msg['vendorid'] == "FEC"
        assert msg['callsign'] == ""
        assert msg['to_bow'] == 0
        assert msg['to_stern'] == 0
        assert msg['to_port'] == 0
        assert msg['to_starboard'] == 0
        assert msg['mothership_mmsi'] == "000000000"

    def test_msg_type_25(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,I6SWo?8P00a3PKpEKEVj0?vNP<65,0*73").decode()

        assert msg['type'] == 25
        assert msg['addressed']
        assert not msg['structured']
        assert msg['dest_mmsi'] == "134218384"

    def test_msg_type_26(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,A,JB3R0GO7p>vQL8tjw0b5hqpd0706kh9d3lR2vbl0400,2*40").decode()
        assert msg['type'] == 26
        assert msg['addressed']
        assert msg['structured']
        assert msg['dest_mmsi'] == "838351848"

        msg = NMEAMessage(b"!AIVDM,1,1,,A,J0@00@370>t0Lh3P0000200H:2rN92,4*14").decode()
        assert msg['type'] == 26
        assert not msg['addressed']
        assert not msg['structured']

    def test_msg_type_27(self):
        msg = NMEAMessage(b"!AIVDM,1,1,,B,KC5E2b@U19PFdLbMuc5=ROv62<7m,0*16").decode(silent=False)
        assert msg
        assert msg['type'] == 27
        assert msg['mmsi'] == "206914217"
        assert msg['accuracy'] == 0
        assert msg['raim'] == 0
        assert msg['status'] == NavigationStatus.NotUnderCommand
        assert round(msg['lon'], 3) == 137.023
        assert round(msg['lat'], 2) == 4.84
        assert msg['speed'] == 57
        assert msg['course'] == 167
        assert msg['gnss'] == 0

    def test_broken_messages(self):
        # Undefined epfd
        assert NMEAMessage(b"!AIVDM,1,1,,B,4>O7m7Iu@<9qUfbtm`vSnwvH20S8,0*46").decode()['epfd'] == EpfdType.Undefined

    def test_multiline_message(self):
        # these messages caused issue #3
        msg_1_part_0 = b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07'
        msg_1_part_1 = b'!AIVDM,2,2,1,A,F@V@00000000000,2*35'

        assert NMEAMessage.assemble_from_iterable(
            messages=[
                NMEAMessage(msg_1_part_0),
                NMEAMessage(msg_1_part_1)
            ]
        ).decode().to_json()

        msg_2_part_0 = b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F'
        msg_2_part_1 = b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D'

        assert NMEAMessage.assemble_from_iterable(
            messages=[
                NMEAMessage(msg_2_part_0),
                NMEAMessage(msg_2_part_1)
            ]
        ).decode().to_json()

    def test_byte_stream(self):
        messages = [
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
            b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
            b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
            b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D',
        ]
        counter = 0
        for msg in ByteStream(messages):
            decoded = msg.decode()
            assert decoded['shipname'] == 'NORDIC HAMBURG'
            assert decoded['mmsi'] == "210035000"
            assert decoded
            counter += 1
        assert counter == 2

    def test_fail_silently(self):
        # this tests combines testing for an UnknownMessageException and the silent param at once
        msg = b"!AIVDM,1,1,,A,U31<0OOP000CshrMdl600?wP00SL,0*43"
        nmea = NMEAMessage(msg)

        with self.assertRaises(UnknownMessageException):
            nmea.decode(silent=False)

        # by default errors are ignored and an empty AIS message is returned
        assert nmea.decode() is not None
        assert isinstance(nmea.decode(), AISMessage)
        text = """{
    "nmea": {
        "ais_id": 37,
        "raw": "!AIVDM,1,1,,A,U31<0OOP000CshrMdl600?wP00SL,0*43",
        "talker": "AI",
        "type": "VDM",
        "message_fragments": 1,
        "fragment_number": 1,
        "message_id": null,
        "channel": "A",
        "payload": "U31<0OOP000CshrMdl600?wP00SL",
        "fill_bits": 0,
        "checksum": 67,
        "bit_array": "100101000011000001001100000000011111011111100000000000000000000000010011111011110000111010011101101100110100000110000000000000001111111111100000000000000000100011011100"
    },
    "decoded": {}
}"""
        self.assertEqual(nmea.decode().to_json(), text)



import socket
import threading
import time
import typing
import unittest

from pyais.stream import OutOfOrderByteStream, UDPStream
from pyais.util import FixedSizeDict
from tests.utils.skip import is_linux
from tests.utils.timeout import time_limit

MESSAGES = [
    b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
    b"!AIVDM,2,1,7,A,543ri001fIOiEa4<0010u84@4000000000000016;hD854o506SRBkk0FAEP,0*07",
    b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
    b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
    b"!AIVDM,2,2,7,A,00000000000,2*23",
    b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D',
]


class MockUDPServer(object):

    def __init__(self, host, port) -> None:
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.host = host
        self.port = port

    def send(self):
        time.sleep(0.1)
        for msg in MESSAGES:
            self.sock.sendto(msg + b"\r\n", (self.host, self.port))

        self.sock.close()


class TestOutOfOrderByteStream(OutOfOrderByteStream):
    """
    Subclass OutOfOrderByteStream to test itÄs message assembly logic without real sockets
    """

    def __init__(self, iterable: typing.Iterable[bytes]) -> None:
        # just accept some messages which then will be used as input
        self.iterable: typing.Iterable[bytes] = iterable
        super().__init__(None)

    def read(self) -> typing.Generator[bytes, None, None]:
        yield from (msg for msg in self.iterable)


class TestOutOfOrder(unittest.TestCase):
    def _spawn_test_server(self):
        self.server = MockUDPServer('127.0.0.1', 9999)
        self.server_thread = threading.Thread(target=self.server.send)
        self.server_thread.start()

    def test_fixed_sized_dict(self):
        N = 10000
        queue = FixedSizeDict(N + 1)
        for i in range(N):
            queue[i] = i

        # no keys were delted
        assert len(queue) == N
        assert queue.popitem(last=False)[0] == 0
        assert queue.popitem(last=True)[0] == N - 1

        # add another
        queue[N + 1] = 35
        queue[N + 2] = 35
        queue[N + 3] = 35
        # now 1/5th of keys is delted
        assert len(queue) == N - (N // 5) + 1
        # make sure only the oldest ones were deleted
        assert queue.popitem(last=False)[0] == (N // 5) + 1

    @unittest.skipIf(not is_linux(), "Skipping because Signal is not available on non unix systems!")
    def test_stream(self):
        # Test the UDP stream with real data
        with time_limit(1):  # make sure the function cannot run forever
            self._spawn_test_server()
            host = "127.0.0.1"
            port = 9999
            counter = 0
            with UDPStream(host, port) as stream:
                for msg in stream:
                    assert msg.decode()
                    counter += 1

                    if counter == 3:
                        break

            self.server_thread.join()

    def test_out_of_order(self):
        messages = [
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
            b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
            b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
            b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D',
        ]
        counter = 0
        for msg in TestOutOfOrderByteStream(messages):
            msg.decode()
            counter += 1
        assert counter == 2

    def test_out_fo_order_in_order(self):
        # in order messages do not cause problems
        messages = [
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
            b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
            b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
            b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D',
        ]
        counter = 0
        for msg in TestOutOfOrderByteStream(messages):
            msg.decode()
            counter += 1
        assert counter == 2

    def test_split_nmea_header_method(self):
        stream = TestOutOfOrderByteStream([])
        msg = b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07'
        stream._split_nmea_header(msg)
        assert stream.seq_id == 1
        assert stream.fragment_offset == 0
        assert stream.fragment_count == 2

        # sequence id could be large
        msg = b'!AIVDM,2,1,145859,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07'
        stream._split_nmea_header(msg)
        assert stream.seq_id == 145859

    def test_index_error(self):
        messages = [
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
            b'!AIVDM,2,9,2,A,F@V@00000000000,2*3D',
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
            b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
        ]
        counter = 0
        for msg in TestOutOfOrderByteStream(messages):
            msg.decode()
            counter += 1

        # only one message was yielded
        assert counter == 1

    def test_delete_after_yield(self):
        messages = [
            b'!AIVDM,2,1,1,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*07',
            b"!AIVDM,2,1,7,A,543ri001fIOiEa4<0010u84@4000000000000016;hD854o506SRBkk0FAEP,0*07",
            b'!AIVDM,2,1,9,A,538CQ>02A;h?D9QC800pu8@T>0P4l9E8L0000017Ah:;;5r50Ahm5;C0,0*0F',
            b'!AIVDM,2,2,1,A,F@V@00000000000,2*35',
            b"!AIVDM,2,2,7,A,00000000000,2*23",
            b'!AIVDM,2,2,9,A,F@V@00000000000,2*3D',
        ]
        stream = TestOutOfOrderByteStream(messages)
        iter_steam = iter(stream)

        # assure that messages are deleted after they are yielded
        assert next(iter_steam).message_id == 1
        assert len(stream._queue) == 2
        assert next(iter_steam).message_id == 7
        assert len(stream._queue) == 1

    def test_three(self):
        messages = [
            b"!AIVDM,3,1,5,A,36KVnDh02wawaHPDA8T8h6tT8000t=AV=maD7?>BWiKIE@TR<2QfvaAF1ST4H31B,0*35",
            b"!AIVDM,3,2,5,A,8IBP:UFW<M0FVWS0DPK19@nh4UdS:OufWUIfPF5l1U9LILBn@9@F:41Q@U1EEOE3,0*1D",
            b"!AIVDM,3,3,5,A,j,0*79"
        ]
        stream = TestOutOfOrderByteStream(messages)
        iter_steam = iter(stream)
        assert next(iter_steam).decode()



from functools import partial
from typing import Any, Dict, Union

import bitarray  # type: ignore

from pyais import messages
from pyais.constants import (
    NavigationStatus,
    ManeuverIndicator,
    TransmitMode,
    EpfdType,
    ShipType,
    StationType,
    StationIntervals,
    NavAid
)
from pyais.exceptions import UnknownMessageException
from pyais.util import get_int, encode_bin_as_ascii6, get_mmsi


def decode_msg_1(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_types_1_2_and_3_position_report_class_a
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'status': NavigationStatus(get_int_from_data(38, 42)),
        'turn': get_int_from_data(42, 50, signed=True),
        'speed': get_int_from_data(50, 60) / 10.0,
        'accuracy': bit_arr[60],
        'lon': get_int_from_data(61, 89, signed=True) / 600000.0,
        'lat': get_int_from_data(89, 116, signed=True) / 600000.0,
        'course': get_int_from_data(116, 128) * 0.1,
        'heading': get_int_from_data(128, 137),
        'second': get_int_from_data(137, 143),
        'maneuver': ManeuverIndicator(get_int_from_data(143, 145)),
        'raim': bit_arr[148],
        'radio': get_int_from_data(149, len(bit_arr)),
    }


def decode_msg_2(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_types_1_2_and_3_position_report_class_a
    """
    return decode_msg_1(bit_arr)


def decode_msg_3(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    AIS Vessel position report using ITDMA (Incremental Time Division Multiple Access)
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_types_1_2_and_3_position_report_class_a
    """
    return decode_msg_1(bit_arr)


def decode_msg_4(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    AIS Vessel position report using SOTDMA (Self-Organizing Time Division Multiple Access)
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_4_base_station_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'year': get_int_from_data(38, 52),
        'month': get_int_from_data(52, 56),
        'day': get_int_from_data(56, 61),
        'hour': get_int_from_data(61, 66),
        'minute': get_int_from_data(66, 72),
        'second': get_int_from_data(72, 78),
        'accuracy': bit_arr[78],
        'lon': get_int_from_data(79, 107, signed=True) / 600000.0,
        'lat': get_int_from_data(107, 134, signed=True) / 600000.0,
        'epfd': EpfdType(get_int_from_data(134, 138)),
        'raim': bit_arr[148],
        'radio': get_int_from_data(148, len(bit_arr)),
    }


def decode_msg_5(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Static and Voyage Related Data
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_5_static_and_voyage_related_data
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'ais_version': get_int_from_data(38, 40),
        'imo': get_int_from_data(40, 70),
        'callsign': encode_bin_as_ascii6(bit_arr[70:112]),
        'shipname': encode_bin_as_ascii6(bit_arr[112:232]),
        'shiptype': ShipType(get_int_from_data(232, 240)),
        'to_bow': get_int_from_data(240, 249),
        'to_stern': get_int_from_data(249, 258),
        'to_port': get_int_from_data(258, 264),
        'to_starboard': get_int_from_data(264, 270),
        'epfd': EpfdType(get_int_from_data(270, 274)),
        'month': get_int_from_data(274, 278),
        'day': get_int_from_data(278, 283),
        'hour': get_int_from_data(283, 288),
        'minute': get_int_from_data(288, 294),
        'draught': get_int_from_data(294, 302) / 10.0,
        'destination': encode_bin_as_ascii6(bit_arr[302:422]),
        'dte': bit_arr[-2]
    }


def decode_msg_6(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Binary Addresses Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_4_base_station_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'seqno': get_int_from_data(38, 40),
        'dest_mmsi': get_mmsi(bit_arr, 40, 70),
        'retransmit': bit_arr[70],
        'dac': get_int_from_data(72, 82),
        'fid': get_int_from_data(82, 88),
        'data': bit_arr[88:].to01()
    }


def decode_msg_7(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Binary Acknowledge
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_7_binary_acknowledge
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'mmsi1': get_mmsi(bit_arr, 40, 70),
        'mmsiseq1': get_int_from_data(70, 72),
        'mmsi2': get_mmsi(bit_arr, 72, 102),
        'mmsiseq2': get_int_from_data(102, 104),
        'mmsi3': get_mmsi(bit_arr, 104, 134),
        'mmsiseq3': get_int_from_data(134, 136),
        'mmsi4': get_mmsi(bit_arr, 136, 166),
        'mmsiseq4': get_int_from_data(166, 168)
    }


def decode_msg_8(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Binary Acknowledge
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_8_binary_broadcast_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'dac': get_int_from_data(40, 50),
        'fid': get_int_from_data(50, 56),
        'data': bit_arr[56:].to01()
    }


def decode_msg_9(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Standard SAR Aircraft Position Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_9_standard_sar_aircraft_position_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'alt': get_int_from_data(38, 50),
        'speed': get_int_from_data(50, 60),
        'accuracy': bit_arr[60],
        'lon': get_int_from_data(61, 89, signed=True) / 600000.0,
        'lat': get_int_from_data(89, 116, signed=True) / 600000.0,
        'course': get_int_from_data(116, 128) * 0.1,
        'second': get_int_from_data(128, 134),
        'dte': bit_arr[142],
        'assigned': bit_arr[146],
        'raim': bit_arr[147],
        'radio': get_int_from_data(148, 168)
    }


def decode_msg_10(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    UTC/Date Inquiry
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_10_utc_date_inquiry
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'dest_mmsi': get_mmsi(bit_arr, 40, 70)
    }


def decode_msg_11(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    UTC/Date Response
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_11_utc_date_response
    """
    return decode_msg_4(bit_arr)


def decode_msg_12(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Addressed Safety-Related Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_12_addressed_safety_related_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'seqno': get_int_from_data(38, 40),
        'dest_mmsi': get_mmsi(bit_arr, 40, 70),
        'retransmit': bit_arr[70],
        'text': encode_bin_as_ascii6(bit_arr[72:])
    }


def decode_msg_13(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Identical to type 7
    """
    return decode_msg_7(bit_arr)


def decode_msg_14(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Safety-Related Broadcast Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_14_safety_related_broadcast_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'text': encode_bin_as_ascii6(bit_arr[40:])
    }


def decode_msg_15(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Interrogation
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_15_interrogation
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'mmsi1': get_mmsi(bit_arr, 40, 70),
        'type1_1': get_int_from_data(70, 76),
        'offset1_1': get_int_from_data(76, 88),
        'type1_2': get_int_from_data(90, 96),
        'offset1_2': get_int_from_data(96, 108),
        'mmsi2': get_mmsi(bit_arr, 110, 140),
        'type2_1': get_int_from_data(140, 146),
        'offset2_1': get_int_from_data(146, 157),
    }


def decode_msg_16(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Assignment Mode Command
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_16_assignment_mode_command
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'mmsi1': get_mmsi(bit_arr, 40, 70),
        'offset1': get_int_from_data(70, 82),
        'increment1': get_int_from_data(82, 92),
        'mmsi2': get_mmsi(bit_arr, 92, 122),
        'offset2': get_int_from_data(122, 134),
        'increment2': get_int_from_data(134, 144)
    }


def decode_msg_17(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    DGNSS Broadcast Binary Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_17_dgnss_broadcast_binary_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(6, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'lon': get_int_from_data(40, 58, signed=True),
        'lat': get_int_from_data(58, 75, signed=True),
        'data': get_int_from_data(80, 816)
    }


def decode_msg_18(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Standard Class B CS Position Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_18_standard_class_b_cs_position_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'speed': get_int_from_data(46, 56) * 0.1,
        'accuracy': bit_arr[56],
        'lon': get_int_from_data(57, 85, signed=True) / 600000.0,
        'lat': get_int_from_data(85, 112, signed=True) / 600000.0,
        'course': get_int_from_data(112, 124) * 0.1,
        'heading': get_int_from_data(124, 133),
        'second': get_int_from_data(133, 139),
        'regional': get_int_from_data(139, 141),
        'cs': bit_arr[141],
        'display': bit_arr[142],
        'dsc': bit_arr[143],
        'band': bit_arr[144],
        'msg22': bit_arr[145],
        'assigned': bit_arr[146],
        'raim': bit_arr[147],
        'radio': get_int_from_data(148, len(bit_arr)),
    }


def decode_msg_19(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Extended Class B CS Position Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_19_extended_class_b_cs_position_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'speed': get_int_from_data(46, 56) * 0.1,
        'accuracy': bit_arr[56],
        'lon': get_int_from_data(57, 85, signed=True) / 600000.0,
        'lat': get_int_from_data(85, 112, signed=True) / 600000.0,
        'course': get_int_from_data(112, 124) * 0.1,
        'heading': get_int_from_data(124, 133),
        'second': get_int_from_data(133, 139),
        'regional': get_int_from_data(139, 143),
        'shipname': encode_bin_as_ascii6(bit_arr[143:263]),
        'shiptype': ShipType(get_int_from_data(263, 271)),
        'to_bow': get_int_from_data(271, 280),
        'to_stern': get_int_from_data(280, 289),
        'to_port': get_int_from_data(289, 295),
        'to_starboard': get_int_from_data(295, 301),
        'epfd': EpfdType(get_int_from_data(301, 305)),
        'raim': bit_arr[305],
        'dte': bit_arr[306],
        'assigned': bit_arr[307],
    }


def decode_msg_20(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Data Link Management Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_20_data_link_management_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'offset1': get_int_from_data(40, 52),
        'number1': get_int_from_data(52, 56),
        'timeout1': get_int_from_data(56, 59),
        'increment1': get_int_from_data(59, 70),

        'offset2': get_int_from_data(70, 82),
        'number2': get_int_from_data(82, 86),
        'timeout2': get_int_from_data(86, 89),
        'increment2': get_int_from_data(89, 100),

        'offset3': get_int_from_data(100, 112),
        'number3': get_int_from_data(112, 116),
        'timeout3': get_int_from_data(116, 119),
        'increment3': get_int_from_data(110, 130),

        'offset4': get_int_from_data(130, 142),
        'number4': get_int_from_data(142, 146),
        'timeout4': get_int_from_data(146, 149),
        'increment4': get_int_from_data(149, 160),
    }


def decode_msg_21(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Aid-to-Navigation Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_21_aid_to_navigation_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'aid_type': NavAid(get_int_from_data(38, 43)),
        'name': encode_bin_as_ascii6(bit_arr[43:163]),
        'accuracy': bit_arr[163],

        'lon': get_int_from_data(164, 192, signed=True) / 600000.0,
        'lat': get_int_from_data(192, 219, signed=True) / 600000.0,

        'to_bow': get_int_from_data(219, 228),
        'to_stern': get_int_from_data(228, 237),
        'to_port': get_int_from_data(237, 243),
        'to_starboard': get_int_from_data(243, 249),

        'epfd': EpfdType(get_int_from_data(249, 253)),
        'second': get_int_from_data(253, 259),
        'off_position': bit_arr[259],
        'regional': get_int_from_data(260, 268),
        'raim': bit_arr[268],
        'virtual_aid': bit_arr[269],
        'assigned': bit_arr[270],
        'name_extension': encode_bin_as_ascii6(bit_arr[272:]),
    }


def decode_msg_22(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Channel Management
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_22_channel_management
    """
    get_int_from_data = partial(get_int, bit_arr)
    data = {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'channel_a': get_int_from_data(40, 52),
        'channel_b': get_int_from_data(52, 64),
        'txrx': get_int_from_data(64, 68),
        'power': bit_arr[68],
        'addressed': bit_arr[139],
        'band_a': bit_arr[140],
        'band_b': bit_arr[141],
        'zonesize': get_int_from_data(142, 145),
    }

    # Broadcast
    d: Dict[str, Any] = {}
    if data['addressed']:
        d = {
            'dest1': get_mmsi(bit_arr, 69, 99),
            'dest2': get_mmsi(bit_arr, 104, 134),
        }
    # Addressed
    else:
        d = {
            'ne_lon': get_int_from_data(69, 87, signed=True) * 0.1,
            'ne_lat': get_int_from_data(87, 104, signed=True) * 0.1,
            'sw_lon': get_int_from_data(104, 122, signed=True) * 0.1,
            'sw_lat': get_int_from_data(122, 139, signed=True) * 0.1,
        }

    data.update(d)
    return data


def decode_msg_23(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Group Assignment Command
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_23_group_assignment_command
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'ne_lon': get_int_from_data(40, 58, signed=True) * 0.1,
        'ne_lat': get_int_from_data(58, 75, signed=True) * 0.1,
        'sw_lon': get_int_from_data(75, 93, signed=True) * 0.1,
        'sw_lat': get_int_from_data(93, 110, signed=True) * 0.1,

        'station_type': StationType(get_int_from_data(110, 114)),
        'ship_type': ShipType(get_int_from_data(114, 122)),
        'txrx': TransmitMode(get_int_from_data(144, 146)),
        'interval': StationIntervals(get_int_from_data(146, 150)),
        'quiet': get_int_from_data(150, 154),
    }


def decode_msg_24(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Static Data Report
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_24_static_data_report
    """
    get_int_from_data = partial(get_int, bit_arr)
    data = {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),
        'partno': get_int_from_data(38, 40)
    }

    d: Dict[str, Any]
    if not data['partno']:
        # Part A
        d = {
            'shipname': encode_bin_as_ascii6(bit_arr[40: 160])
        }
    else:
        # Part B
        d = {
            'shiptype': ShipType(get_int_from_data(40, 48)),
            'vendorid': encode_bin_as_ascii6(bit_arr[48: 66]),
            'model': get_int_from_data(66, 70),
            'serial': get_int_from_data(70, 90),
            'callsign': encode_bin_as_ascii6(bit_arr[90: 132]),
            'to_bow': get_int_from_data(132, 141),
            'to_stern': get_int_from_data(141, 150),
            'to_port': get_int_from_data(150, 156),
            'to_starboard': get_int_from_data(156, 162),
            'mothership_mmsi': get_mmsi(bit_arr, 132, 162)
        }
    data.update(d)
    return data


def decode_msg_25(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Single Slot Binary Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_25_single_slot_binary_message

    NOTE: This message type is quite uncommon and
    I was not able find any real world occurrence of the type.
    Also documentation seems to vary. Use with caution.
    """
    get_int_from_data = partial(get_int, bit_arr)
    data: Dict[str, Any] = {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'addressed': bit_arr[38],
        'structured': bit_arr[39],
    }

    d: Dict[str, Any]
    if data['addressed']:
        d = {
            'dest_mmsi': get_mmsi(bit_arr, 40, 70),
        }
        data.update(d)

    lo_ix = 40 if data['addressed'] else 70
    hi_ix = lo_ix + 16

    if data['structured']:
        d = {
            'app_id': get_int_from_data(lo_ix, hi_ix),
            'data': bit_arr[hi_ix:].to01()
        }
    else:
        d = {
            'data': bit_arr[lo_ix:].to01()
        }
    data.update(d)
    return data


def decode_msg_26(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Multiple Slot Binary Message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_26_multiple_slot_binary_message

    NOTE: This message type is quite uncommon and
    I was not able find any real world occurrence of the type.
    Also documentation seems to vary. Use with caution.
    """
    get_int_from_data = partial(get_int, bit_arr)
    radio_status_offset = len(bit_arr) - 20

    data = {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'addressed': bit_arr[38],
        'structured': bit_arr[39],
        'radio': get_int_from_data(radio_status_offset, len(bit_arr))
    }

    d: Dict[str, Any]
    if data['addressed']:
        d = {
            'dest_mmsi': get_mmsi(bit_arr, 40, 70),
        }
        data.update(d)

    lo_ix = 40 if data['addressed'] else 70
    hi_ix = lo_ix + 16

    if data['structured']:
        d = {
            'app_id': get_int_from_data(lo_ix, hi_ix),
            'data': bit_arr[hi_ix:radio_status_offset].to01()
        }
    else:
        d = {
            'data': bit_arr[lo_ix:radio_status_offset].to01()
        }

    data.update(d)
    return data


def decode_msg_27(bit_arr: bitarray.bitarray) -> Dict[str, Any]:
    """
    Long Range AIS Broadcast message
    Src: https://gpsd.gitlab.io/gpsd/AIVDM.html#_type_27_long_range_ais_broadcast_message
    """
    get_int_from_data = partial(get_int, bit_arr)
    return {
        'type': get_int_from_data(0, 6),
        'repeat': get_int_from_data(8, 8),
        'mmsi': get_mmsi(bit_arr, 8, 38),

        'accuracy': bit_arr[38],
        'raim': bit_arr[39],
        'status': NavigationStatus(get_int_from_data(40, 44)),
        'lon': get_int_from_data(44, 62, signed=True) / 600.0,
        'lat': get_int_from_data(62, 79, signed=True) / 600.0,
        'speed': get_int_from_data(79, 85),
        'course': get_int_from_data(85, 94),
        'gnss': bit_arr[94],
    }


# Decoding Lookup Table
DECODE_MSG = [
    decode_msg_1,  # there are messages with a zero (0) as an id. these seem to be the same as type 1 messages
    decode_msg_1,
    decode_msg_2,
    decode_msg_3,
    decode_msg_4,
    decode_msg_5,
    decode_msg_6,
    decode_msg_7,
    decode_msg_8,
    decode_msg_9,
    decode_msg_10,
    decode_msg_11,
    decode_msg_12,
    decode_msg_13,
    decode_msg_14,
    decode_msg_15,
    decode_msg_16,
    decode_msg_17,
    decode_msg_18,
    decode_msg_19,
    decode_msg_20,
    decode_msg_21,
    decode_msg_22,
    decode_msg_23,
    decode_msg_24,
    decode_msg_25,
    decode_msg_26,
    decode_msg_27,
]


def _decode(msg: "messages.NMEAMessage") -> Dict[str, Any]:
    """
    Decodes a given NMEA message.
    """
    try:
        return DECODE_MSG[msg.ais_id](msg.bit_array)
    except IndexError as e:
        raise UnknownMessageException(f"The message {msg} is not currently supported!") from e


def decode(msg: "messages.NMEAMessage") -> Dict[str, Any]:
    """
    Decodes a given message.

    @param msg: A object of type NMEAMessage to decode
    """
    return _decode(msg)


def decode_msg(*args: Union[str, bytes]) -> Dict[str, Any]:
    """
    Decode single message.

    This method is ONLY meant to decode a SINGLE (multiline) message.
    Pass every part of a single as an argument.

    @param args: A AIS message, that can be either bytes or str (UTF-8) encoded.
                For multiline messages, pass all parts.
    @return: A dictionary of the decoded key-value pairs.

    """
    # Make everything bytes
    message_as_bytes = tuple(msg.encode('utf-8') if isinstance(msg, str) else msg for msg in args)

    # Create temporary messages
    temp = [messages.NMEAMessage(m) for m in message_as_bytes]

    # Assemble temporary messages
    final = messages.NMEAMessage.assemble_from_iterable(temp)

    # Decode
    return final.decode(silent=False).content  # type: ignore



import unittest
from pyais.constants import NavigationStatus, ManeuverIndicator, ShipType, NavAid, TransmitMode, StationIntervals, \
    StationType


class TestConstants(unittest.TestCase):
    def test_nav_status(self):
        self.assertEqual(NavigationStatus(3), NavigationStatus.RestrictedManoeuverability)
        self.assertEqual(NavigationStatus(17), NavigationStatus.Undefined)

    def test_maneuver_indication(self):
        self.assertEqual(ManeuverIndicator(0), ManeuverIndicator.NotAvailable)
        self.assertEqual(ManeuverIndicator(1), ManeuverIndicator.NoSpecialManeuver)
        self.assertEqual(ManeuverIndicator(2), ManeuverIndicator.SpecialManeuver)
        self.assertEqual(ManeuverIndicator(3), ManeuverIndicator.UNDEFINED)
        self.assertEqual(ManeuverIndicator(4), ManeuverIndicator.UNDEFINED)

    def test_ship_types(self):
        self.assertEqual(ShipType(25), ShipType.WIG_Reserved)
        self.assertEqual(ShipType(46), ShipType.HSC_Reserved)
        self.assertEqual(ShipType(57), ShipType.SPARE)
        self.assertEqual(ShipType(68), ShipType.Passenger_Reserved)
        self.assertEqual(ShipType(78), ShipType.Cargo_Reserved)
        self.assertEqual(ShipType(85), ShipType.Tanker_Reserved)
        self.assertEqual(ShipType(96), ShipType.OtherType_Reserved)
        self.assertEqual(ShipType(100), ShipType.NotAvailable)

    def test_navaid(self):
        self.assertEqual(NavAid(23), NavAid.CARDINAL_MARK_W)
        self.assertEqual(NavAid(32), NavAid.DEFAULT)

    def test_t_mode(self):
        self.assertEqual(TransmitMode(0), TransmitMode.TXA_TXB_RXA_RXB)
        self.assertEqual(TransmitMode(10), TransmitMode.TXA_TXB_RXA_RXB)

    def test_station_types(self):
        self.assertEqual(StationType(0), StationType.ALL)
        self.assertEqual(StationType(11), StationType.RESERVED)
        self.assertEqual(StationType(7), StationType.REGIONAL)
        self.assertEqual(StationType(16), StationType.ALL)

    def test_station_intervals(self):
        self.assertEqual(StationIntervals(0), StationIntervals.AUTONOMOUS_MODE)
        self.assertEqual(StationIntervals(11), StationIntervals.RESERVED)
        self.assertEqual(StationIntervals(7), StationIntervals.SECONDS_10)
        self.assertEqual(StationIntervals(12), StationIntervals.RESERVED)



import os

import setuptools  # type:ignore

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(os.path.join('pyais', '__init__.py')) as f:
    for line in f:
        if line.strip().startswith('__version__'):
            VERSION = line.split('=')[1].strip()[1:-1].strip()
            break

setuptools.setup(
    name="pyais",
    version=VERSION,
    author="Leon Morten Richter",
    author_email="leon.morten@gmail.com",
    description="Ais message decoding",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/M0r13n/pyais",
    license="MIT",
    packages=setuptools.find_packages(),
    package_data={
        "pyais": ["py.typed"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Communications",
        "Topic :: System :: Networking",
        "Typing :: Typed",
    ],
    keywords=["AIS", "ship", "decoding", "nmea"],
    python_requires='>=3.6',
    install_requires=[
        "bitarray"
    ],
    entry_points={
        "console_scripts": [
            'ais-decode=pyais.main:main'
        ]
    }
)



from pyais.messages import NMEAMessage, AISMessage
from pyais.stream import TCPStream, FileReaderStream, IterMessages
from pyais.decode import decode_msg


__license__ = 'MIT'
__version__ = '1.6.0'

__all__ = (
    'decode_msg',
    'NMEAMessage',
    'AISMessage',
    'TCPStream',
    'IterMessages',
    'FileReaderStream'
)
